#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libguile.h>
#include <expat.h>

#if  (SCM_MAJOR_VERSION  == 1) && (SCM_MINOR_VERSION == 6)
SCM scm_from_locale_stringn(const char *str,int n)
{
  return scm_mem2string(str,n);
}

SCM scm_from_locale_string(const char *str)
{
  return scm_from_locale_stringn(str?str:"",str?strlen(str):0);
}

SCM scm_from_locale_symbol(const char *str)
{
  return scm_str2symbol(str);
}
#endif

static void XMLCALL
element_start(void *data, const char *el, const char **attr)
{
  SCM *tokens = (SCM*)(data);
  SCM answer = SCM_EOL;

  answer = scm_cons(scm_from_locale_symbol(el),answer);
  for (;*attr;attr+=2)
    {
      const char *name = *attr;
      const char *value = *(attr + 1);
      answer = scm_cons(scm_from_locale_symbol(name),
			answer);
      answer = scm_cons(scm_from_locale_stringn(value,strlen(value)),
			answer);
    }
	       

 *tokens = scm_cons(answer,*tokens);
}

static void XMLCALL
element_stop(void *data, const char *el)
{
  SCM *tokens = (SCM*)(data);
  SCM answer = SCM_EOL;
  
  answer = scm_cons(scm_from_locale_symbol("/"),answer);
  answer = scm_cons(scm_from_locale_symbol(el),answer);
  
  *tokens = scm_cons(answer,*tokens);
}


static void XMLCALL
comment_handler(void *data,const char *buffer)
{
  SCM *tokens = (SCM*)(data);

  SCM comment = scm_from_locale_symbol("!--");
  SCM end = scm_from_locale_symbol("/");
  
  *tokens = scm_cons(scm_cons(comment,SCM_EOL),*tokens);
  *tokens = scm_cons(scm_from_locale_string(buffer),*tokens);
  *tokens = scm_cons(scm_cons(comment,scm_cons(end,SCM_EOL)),*tokens);
}


static void XMLCALL
default_handler(void *data,const char *buffer,int len)
{
  SCM *tokens = (SCM*)(data);
  *tokens = scm_cons(scm_from_locale_stringn(buffer,len),
		     *tokens);
}

static void XMLCALL
cdata_handler(void *data,const char *buffer, int len)
{
  default_handler(data,buffer,len);
}

static void XMLCALL
doctype_start(void* data,const char* name,const char* sysid,const char *pubid,int has_subset)
{ /*nothing*/ }

static void XMLCALL
doctype_stop(void* data)
{ /*nothing*/ }

static void XMLCALL
skipped_entity_handler(void *data,const char *name,int is_param)
{
  SCM *tokens = (SCM*)(data);

  SCM entity = scm_from_locale_symbol("&");
  SCM end = scm_from_locale_symbol("/");
  
  *tokens = scm_cons(scm_cons(entity,SCM_EOL),*tokens);
  *tokens = scm_cons(scm_from_locale_string(name),*tokens);
  *tokens = scm_cons(scm_cons(entity,scm_cons(end,SCM_EOL)),*tokens);
}

SCM
scm_xml_tokenize(SCM port)
{
  SCM_ASSERT(SCM_NFALSEP(scm_input_port_p(port)),
	     port, SCM_ARG1, "scm_xml_tokenize");

  char buffer[BUFSIZ];
  size_t len;
  XML_Parser parser = XML_ParserCreate(NULL);
  SCM tokens = SCM_EOL;

  if (!parser) {
    fprintf(stderr,"Couldn't allocate memory for parser\n");
    return SCM_BOOL_F;
  }

  XML_SetElementHandler(parser, element_start, element_stop);
  XML_SetCommentHandler(parser,comment_handler);
  XML_SetCharacterDataHandler(parser,cdata_handler);//pre-defined entities will go here
  XML_SetDoctypeDeclHandler(parser, doctype_start, doctype_stop);//need to allow users use external entities
  XML_SetSkippedEntityHandler(parser, skipped_entity_handler);//catch skipped html entities
  
  XML_SetUserData(parser, &tokens);
  XML_SetDefaultHandler(parser,default_handler);

  while ((len = scm_c_read(port,buffer,(sizeof buffer) - 1)) > 0)
    {
      if (XML_Parse(parser, buffer, len, XML_FALSE) == XML_STATUS_ERROR)
	{
	  SCM description = SCM_EOL;
	  char *buffer;
	  int result=
	    asprintf(&buffer,"Parse error at line %lu:%s",
		     XML_GetCurrentLineNumber(parser),
		     XML_ErrorString(XML_GetErrorCode(parser)));
	  if (result >= 0)
	    {
	      description = scm_list_1(scm_from_locale_string(buffer)); 
	      free(buffer);
	    }
	  
	  scm_throw(scm_from_locale_symbol("xml-error"),description);
	  return SCM_BOOL_F;
	}
    }
   XML_Parse(parser, buffer, 0, XML_TRUE);
   XML_ParserFree(parser);

  return tokens;
}

void scm_init_xmltokenizer()
{
  scm_c_define_gsubr("xml-tokenize", 1, 0, 0, (SCM (*)())scm_xml_tokenize);
}
